/* 
 * File:   main03c.c
 *
 * Created on 2015/05/05, 17:43
 */

#include <18F2550.h>

#device ansi
#device const=read_only
#device PASS_STRINGS=IN_RAM
#opt compress

// 10MHz -> (PLL prescaler : PLLDIV=3) -> 3.3MHz(=10/3) -> (96MPLL) -> 79MHz(=3.3*24) -> (PLL postscaler : CPUDIV=2) -> 39.5MHz(=79/2)
#fuses PLL3,CPUDIV2,NOUSBDIV,HSPLL,NOFCMEN,NOIESO,NOBROWNOUT,NOVREGEN,NOWDT,CCP2B3,NOPBADEN,NOLPT1OSC,NOMCLR,STVREN,NOLVP,NOXINST,NOCPB,NOCPD
#use delay (clock=39500000)

#bit USBEN = 0x0F6D.3       // PIC18F2550 reference manual P68
#bit UTRDIS = 0x0F6F.3      //

#include <stdio.h>
#include <stdlib.h>

// timer
#define TMRREG_START        3817        // 0.05s
//#define TMRREG_START      53036       // 0.01s
#define CHATTERING_ADJUST   10

#define MP3_SI              PIN_B2
#define MP3_SCLK            PIN_B3
#define MP3_CS              PIN_B4
#define MP3_BSYNC           PIN_C6
#define MP3_DCLK            PIN_C2
#define MP3_DREQ            PIN_C4
#define MP3_RS              PIN_B5

#define SW1                 input(PIN_E3)      //PORTEbits.RE3
#define SW2                 input(PIN_A0)      //PORTAbits.RA0

#define LED_ON              output_high(PIN_A3)
#define LED_OFF             output_low(PIN_A3)

#define MP3_BUFFER_SIZE     512           // 512=32*16
#define SEND_BUF_CNT_UNIT   200           // 512*200=102400=0.1MByte
#define SEND_SIZE           32

#define SW1_PRESSED         0x81
#define SW2_PRESSED         0x82

#define PLAY_NEXT_MP3       0x91
#define PLAY_OTHER_DIR      0x92

#define MIN_VOLUME          0xD0
#define MAX_VOLUME          0x00

#define DIRLISTFILE         "/DIRLIST.TXT"      // directory list
#define HISTFILE            "/HIST.DAT"         // play history
#define NUMDIRSFILE         "/NUMDIRS.DAT"      // number of directories
#define HIST_SIZE           64

#include "integer.h"
#include "diskio.h"
#include "ffconf.h"
#include "ff.h"
#include "user.h"

static FATFS        Fatfs;
static FIL          Fil;        // Attention! (common usage variable)

static uint8_t      VOLUME;
static uint16_t     ST_AMP, ST_FREQ, SB_AMP, SB_FREQ;   // bass/treble control values
static uint8_t      TREBLE_CONTROL;
static uint8_t      MP3DATA[MP3_BUFFER_SIZE];
static uint16_t     SrandSeed;
static uint16_t     Hist[HIST_SIZE];

// bush button control
static BOOLEAN      EVENT_SW1;
static BOOLEAN      EVENT_SW2;
static uint16_t     CHATTERING_TMR;
static uint16_t     SW_CONTINUE_COUNT;
static int          Sw1Last;

static uint8_t      LED_BLINK_INTERVAL;
static BOOLEAN      LED_STATE;
static uint16_t     BlinkTmrCount;


void InitApp(void);
uint8_t PauseAndSelectItem();
uint8_t play_random();
uint8_t play_subdirectory(TCHAR *path, uint16_t dirno);
uint8_t play_mp3(TCHAR *mp3file);
uint16_t read_numdirs_file();
void save_numdirs_file(uint16_t num);
uint8_t check_play_history(uint16_t no);
void add_history(uint16_t no);
uint16_t read_eeprom_uint16(uint8_t address);
void write_eeprom_uint16(uint8_t address, uint16_t data);
void read_histfile();
uint8_t isMp3File(char *fname);
uint16_t write_dirlistfile(char *path);
uint16_t write_dirlistfile_level1(char *path);
uint16_t write_dirlistfile_level2(char *path);

#include "spi_sd.c"
#include "ff.c"
#include "vs.c"

/*
 * 
 */
int main(int argc, char** argv) {
    FRESULT res;

    InitApp();
    VOLUME = MAX_VOLUME;
    ST_AMP = 0x00; ST_FREQ = 0x00; SB_AMP = 0x00; SB_FREQ = 0x00;
    Sw1Last = SW1;

    // Start VS1033
    output_low(MP3_RS);
    wait_ms(100);
    output_high(MP3_RS);
    wait_ms(100);
    //INTCONbits.TMR0IE = 1;      // Enables the TMR0 overflow interrupt

    res = f_mount(&Fatfs, _T("0:"), 0);
    if (res == FR_OK){
        uint8_t     sel;
        sel = PauseAndSelectItem();
        if (sel == SW1_PRESSED){
            uint16_t    seed;
            seed = read_eeprom_uint16(0);
            if (seed == 0xFFFF) seed = 0;
            else seed++;
            seed ^= SrandSeed;
            write_eeprom_uint16(0, seed);
            srand(seed);

            play_random();
        }
    }

    // Error
    do {
        wait_ms(250);
        LED_STATE = 1 - LED_STATE;
        if (LED_STATE == 1) LED_ON;
        else LED_OFF;
    } while (1);
    return (0);
}

void InitApp(void){
    /* Initialize peripherals */
    setup_adc(ADC_OFF);         // ADC disabled
    setup_ccp1(CCP_OFF);        // CCP1 disabled
    setup_ccp2(CCP_OFF);        // CCP2 disabled
    USBEN = 0;      // USB module disabled
    UTRDIS = 1;     // On chip tranciever disabled
    //T1CONbits.T1OSCEN = 0;      // Timer1 oscillator shut off
    //RCSTAbits.SPEN = 0;         // Serial port diabled

    /* Setup analog functionality and port direction */
    set_tris_a(0x01);       // A0=input
    set_tris_b(0x00);       //
    set_tris_c(0x10);       // C4=input 0001 0000
    port_b_pullups(TRUE);

    /* Enable interrupts */
    setup_timer_1(T1_INTERNAL | T1_DIV_BY_8);
    set_timer1(TMRREG_START);
    clear_interrupt(INT_TIMER1);
    enable_interrupts(INT_TIMER1);

    enable_interrupts(GLOBAL);
}


uint8_t PauseAndSelectItem(){
    uint8_t     ret;

    ret = 0;
    LED_BLINK_INTERVAL = 10;
    BlinkTmrCount = 0;
    LED_STATE = 0;

    do {
        if (EVENT_SW1 && CHATTERING_TMR == 0){
            ret = SW1_PRESSED;
            EVENT_SW1 = 0;
            break;
        }
        if (EVENT_SW2 && CHATTERING_TMR == 0){
            ret = SW2_PRESSED;
            EVENT_SW2 = 0;
            break;
        }
        if (SrandSeed == 0xFFFF){
            SrandSeed = 0;
        } else {
            SrandSeed++;
        }
    } while (1);

    LED_BLINK_INTERVAL = 0;
    return ret;
}

#INT_TIMER1
void isr_timer(){
    WORD n;

    n = Timer1;
    if (n) Timer1 = --n;
    n = Timer2;
    if (n) Timer2 = --n;

    if (CHATTERING_TMR > 0){
        CHATTERING_TMR++;
        if (CHATTERING_TMR == CHATTERING_ADJUST){
            CHATTERING_TMR = 0;
        }
    } else {
        uint16_t    Sw1Current;

        // SW1 supports long time press
        Sw1Current = SW1;
        if (Sw1Last == Sw1Current){
            if (Sw1Current == 0){   // press continues
                SW_CONTINUE_COUNT++;
            } else {    // release continues
                SW_CONTINUE_COUNT = 0;
            }
        } else {    // press state change (press or release)
            CHATTERING_TMR = 1;     // chattering timer start
            if (Sw1Current == 1){   // released
                EVENT_SW1 = 1;      // press event happens
            }
        }
        Sw1Last = Sw1Current;

        if (SW2 == 0 && EVENT_SW2 == 0){
            EVENT_SW2 = 1;          // SW2 is ON
            CHATTERING_TMR = 1;     // chattering timer start
        }
    }

    if (LED_BLINK_INTERVAL){
        BlinkTmrCount++;
        if (BlinkTmrCount == LED_BLINK_INTERVAL){
            BlinkTmrCount = 0;
            LED_STATE = ~LED_STATE;
            if (LED_STATE == 1) LED_ON;
            else LED_OFF;
        }
    }
}

uint8_t play_random(){
    FRESULT     res;
//    FIL         Fil;
    uint8_t     ret;
    uint8_t     found;
    TCHAR       buf[40];
    uint16_t    numdirs, targetno, countdirs;
    char        *c;

    ret = 0;
    read_histfile();
    numdirs = read_numdirs_file();

    do {
        targetno = rand() % numdirs;

        countdirs = 0;
        res = f_open(&Fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
        if (res == FR_OK){
            found = 0;
            while (1){
                //if (f_eof(&Fil) == 1){  // EOF OF FILE
                //    break;
                //}
                if (f_gets(buf, 64, &Fil) == NULL){
                    break;
                }
                
                if (countdirs++ == targetno){
                    found = 1;
                    c = strchr(buf, '\r');
                    if (c != NULL) *c = '\0';
                    else {
                        c = strchr(buf, '\n');
                        if (c != NULL) *c = '\0';
                    }
                    break;
                }
            }
            f_close(&Fil);

            if (found == 1){
                if (check_play_history(targetno) == 0){
                    LED_ON;
                    play_subdirectory(buf, targetno);
                    LED_OFF;
                }
            } else if (countdirs != numdirs){
                if (countdirs == 0) return 0;
                numdirs = countdirs;
                save_numdirs_file(numdirs);
            }
        } else {
            numdirs = write_dirlistfile("/");
            if (numdirs == 0) break;
            save_numdirs_file(numdirs);
        }
    } while (1);

    return ret;
}

uint8_t play_subdirectory(TCHAR *path, uint16_t dirno){
    DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t ret;
    uint8_t trackno;

    trackno = 0;
    ret = 0;
    res1 = f_opendir(&dir, path);
    if (res1 != FR_OK){
        return res1;
    }
    do {
        res1 = f_readdir(&dir, &fi);
        if (res1 != FR_OK) break;
        if (fi.fname[0]){       // MP3 name
            if ((fi.fattrib & AM_DIR) == 0 && isMp3File(fi.fname)){
                TCHAR   mp3path[64];

                trackno++;
                if (trackno == 1){
                    add_history(dirno);
                }
                strcpy(mp3path, path);
                strcat(mp3path, "/");
                strcat(mp3path, fi.fname);
                res1 = play_mp3(mp3path);
                if (res1 == 0){
                    // normal end -> next mp3
                } else if (res1 == PLAY_NEXT_MP3){
                } else if (res1 == PLAY_OTHER_DIR){
                    ret = PLAY_OTHER_DIR;
                    break;
                } else {
                    // error
                    ret = res1;
                    break;
                }
            }
        } else {
            break;
        }
    } while (1);
    f_closedir(&dir);

    return ret;
}

uint8_t play_mp3(TCHAR *mp3file){
    FRESULT     res;
//    FIL         Fil;
    UINT        br;
    int         continue_flag, ret;

    ret = 0;
    res = f_open(&Fil, mp3file, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        //if (res == FR_NO_FILE){
        //} else {
        //}
        return res;
    }

    VS1033RESET();
    continue_flag = 1;

    while (continue_flag) {
        res = f_read(&Fil, MP3DATA, MP3_BUFFER_SIZE, &br);
        if (res != FR_OK || br == 0) break;

        VS1033SENDDATA(MP3DATA, br);

        if (br < MP3_BUFFER_SIZE){
            continue_flag = 0;
        } else if (EVENT_SW1 && CHATTERING_TMR == 0){
            EVENT_SW1 = 0;
            if (SW_CONTINUE_COUNT == 0){
                ret = PLAY_NEXT_MP3;
            } else {    // long push
                ret = PLAY_OTHER_DIR;
            }
            SW_CONTINUE_COUNT = 0;
            continue_flag = 0;
        } else if (EVENT_SW2 && CHATTERING_TMR == 0){      //
            EVENT_SW2 = 0;
            TREBLE_CONTROL++;
            if (TREBLE_CONTROL == 1){
                ST_AMP = 0x09;
                ST_FREQ = 0x03;
            /*
            } else if (TREBLE_CONTROL == 2){
                ST_AMP = 0x09;
                ST_FREQ = 0x02;
            } else if (TREBLE_CONTROL == 3){
                ST_AMP = 0x09;
                ST_FREQ = 0x01;
            */
            } else {
                TREBLE_CONTROL = 0;
                ST_AMP = 0x00;
                ST_FREQ = 0x00;
            }
            VS1033COMM(2, ((ST_AMP << 12) | (ST_FREQ << 8) | (SB_AMP << 4) | SB_FREQ));
        }
    }

    f_close(&Fil);
    return ret;
}

uint16_t write_dirlistfile(char *path){
    FRESULT     res;
    DIR         dir;
    FILINFO     fi;
    uint16_t    num;

    num = 0;
    res = f_open(&Fil, DIRLISTFILE, FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
        res = f_opendir(&dir, path);
        if (res == FR_OK){
            while (1){
                res = f_readdir(&dir, &fi);
                if (res != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
                        char subpath[15];
                        strcpy(subpath, path);
                        strcat(subpath, fi.fname);
                        f_puts(subpath, &Fil);
                        f_puts("\n", &Fil);
                        num++;

                        num += write_dirlistfile_level1(subpath);
                    }
                } else break;
            }
            f_closedir(&dir);
        }
        
        f_close(&Fil);
    }
    return num;
}

uint16_t write_dirlistfile_level1(char *path){
    FRESULT     res;
    DIR         dir;
    FILINFO     fi;
    uint16_t    num;

    num = 0;
    res = f_opendir(&dir, path);
    if (res == FR_OK){
        while (1){
            res = f_readdir(&dir, &fi);
            if (res != FR_OK) break;
            if (fi.fname[0]){
                if ((fi.fattrib & AM_DIR) != 0){
                    char subpath[30];
                    strcpy(subpath, path);
                    strcat(subpath, "/");
                    strcat(subpath, fi.fname);
                    f_puts(subpath, &Fil);
                    f_puts("\n", &Fil);
                    num++;

                    num += write_dirlistfile_level2(subpath);
                }
            } else break;
        }
        f_closedir(&dir);
    }
    return num;
}

uint16_t write_dirlistfile_level2(char *path){
    FRESULT     res;
    DIR         dir;
    FILINFO     fi;
    uint16_t    num;

    num = 0;
    res = f_opendir(&dir, path);
    if (res == FR_OK){
        while (1){
            res = f_readdir(&dir, &fi);
            if (res != FR_OK) break;
            if (fi.fname[0]){
                if ((fi.fattrib & AM_DIR) != 0){
                    char subpath[45];
                    strcpy(subpath, path);
                    strcat(subpath, "/");
                    strcat(subpath, fi.fname);
                    f_puts(subpath, &Fil);
                    f_puts("\n", &Fil);
                    num++;

                    //write_dirlistfile_level3(subpath);
                }
            } else break;
        }
        f_closedir(&dir);
    }
    return num;
}

// Get number of total directories
uint16_t read_numdirs_file(){
//    FIL         Fil;
    FRESULT     res;
    uint16_t    num;
    uint16_t    br;
    // NUMDIRSFILE
    res = f_open(&Fil, NUMDIRSFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(&Fil, &num, 2, &br);
        if (res != FR_OK || br == 0 || num == 0) num = 0xFFFF;
        f_close(&Fil);
        return num;
    } else {
        return 0xFFFF;
    }
}

void save_numdirs_file(uint16_t num){
//    FIL         Fil;
    FRESULT     res;
    uint16_t    br;
    res = f_open(&Fil, NUMDIRSFILE, FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
        f_write(&Fil, &num, 2, &br);
        f_close(&Fil);
    }
}

uint8_t check_play_history(uint16_t no){
    int16     i;

    for (i=0; i<HIST_SIZE; i++){
        if (Hist[i] == no) return 1;
    }
    return 0;
}

/*
 * Add folder no to history list and save to file
 */
void add_history(uint16_t no){
    int16       i;
    FRESULT     res;
    uint16_t    br;

    for (i=(HIST_SIZE - 1); i>0; i--){
        Hist[i] = Hist[i-1];
    }
    Hist[0] = no;

    res = f_open(&Fil, HISTFILE, FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
        f_write(&Fil, Hist, 2*HIST_SIZE, &br);
        f_close(&Fil);
    }
}

void read_histfile(){
    FRESULT     res;
    uint16_t    br;

    res = f_open(&Fil, HISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(&Fil, Hist, 2*HIST_SIZE, &br);
        f_close(&Fil);
    }
}

uint16_t read_eeprom_uint16(uint8_t address){
    uint16_t    h;
    h = read_eeprom(address + 1);       // PCH built-in function
    h <<= 8;
    h |= read_eeprom(address);

    return h;
}

void write_eeprom_uint16(uint8_t address, uint16_t data){
    write_eeprom(address, data & 0xFF);
    write_eeprom(address + 1, (data >> 8) & 0xFF);
}

/*
 *  Checks file extension .MP3
 */
uint8_t isMp3File(char *fname){
    char    *extension;
    extension = strchr(fname, '.');
    if (extension != NULL && strcmp(extension, ".MP3") == 0){
        return 1;
    } else {
        return 0;
    }
}

