/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for FatFs     (C)ChaN, 2013        */
/*-----------------------------------------------------------------------*/
/* If a working storage control module is available, it should be        */
/* attached to the FatFs via a glue function rather than modifying it.   */
/* This is an example of glue functions to attach various exsisting      */
/* storage control module to the FatFs module with a defined API.        */
/*-----------------------------------------------------------------------*/
/* Import to PCH.
	1. BYTE -> FFBYTE (defined in Integer.h)
*/

#if defined(__XC)
    #include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>        /* HiTech General Include File */
#elif defined(__18CXX)
    #include <p18cxxx.h>    /* C18 General Include File */
#endif
#include "diskio.h"		/* FatFs lower layer API */
#include "ffconf.h"

/* MMC/SD command */
#define CMD0	(0)		/* GO_IDLE_STATE */
#define CMD1	(1)		/* SEND_OP_COND (MMC) */
#define	ACMD41	(0x80+41)	/* SEND_OP_COND (SDC) */
#define CMD8	(8)		/* SEND_IF_COND */
#define CMD9	(9)		/* SEND_CSD */
#define CMD10	(10)		/* SEND_CID */
#define CMD12	(12)		/* STOP_TRANSMISSION */
#define ACMD13	(0x80+13)	/* SD_STATUS (SDC) */
#define CMD16	(16)		/* SET_BLOCKLEN */
#define CMD17	(17)		/* READ_SINGLE_BLOCK */
#define CMD18	(18)		/* READ_MULTIPLE_BLOCK */
#define CMD23	(23)		/* SET_BLOCK_COUNT (MMC) */
#define	ACMD23	(0x80+23)	/* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24	(24)		/* WRITE_BLOCK */
#define CMD25	(25)		/* WRITE_MULTIPLE_BLOCK */
#define CMD32	(32)		/* ERASE_ER_BLK_START */
#define CMD33	(33)		/* ERASE_ER_BLK_END */
#define CMD38	(38)		/* ERASE */
#define CMD55	(55)		/* APP_CMD */
#define CMD58	(58)		/* READ_OCR */

/* PCH */
#define	CS_LOW()	output_low(PIN_A5)      /* Set CS Low */
#define	CS_HIGH()	output_high(PIN_A5)     /* Set CS High */
//#define	CS_LOW()	PORTAbits.RA5 = 0		/* Set CS Low */
//#define	CS_HIGH()	PORTAbits.RA5 = 1		/* Set CS High */

static volatile
DSTATUS Stat = STA_NOINIT;	/* Physical drive status */

volatile
WORD Timer1, Timer2;	/* 1000Hz decrement timer stopped at zero (driven by disk_timerproc()) */

static
FFBYTE CardType;			/* Card type flags */


/*-----------------------------------------------------------------------*/
/* Inidialize SPI                                                        */
/*-----------------------------------------------------------------------*/
static void
power_on(){
    int16   tris;

  	// PCH
    setup_spi(SPI_MASTER | SPI_CLK_DIV_4 | SPI_SCK_IDLE_LOW | SPI_XMIT_L_TO_H | SPI_SAMPLE_AT_MIDDLE);
    /* XC8
    SSPCON1bits.SSPEN = 1;      // MSSP enabled
//    SSPCON1bits.SSPM = 2;       // SPI master mode, clock=Fosc/64
    SSPCON1bits.SSPM = 0;       // SPI master mode, clock=Fosc/4
    SSPCON1bits.CKP = 0;        // idle state for clock is low level
    SSPSTATbits.CKE = 1;        // 1 = Transmit occurs on transition from active to Idle clock state
    SSPSTATbits.SMP = 0;        //
    */

    // RA5 (SS->SD/CS)
    tris = get_tris_a();
    tris &= ~0x20;   // 0010 0000
    set_tris_a(tris);
    //TRISA5 = 0;     // output (XC8)

    // RB0 (SDI->SD/DO)
    // RB1 (SCK->SD/CLOCK)
    tris = get_tris_b();
    tris |= 0x01;
    tris &= ~0x02;  // 0000 0010
    set_tris_b(tris);
    //TRISB0 = 1;     // input (XC8)
    //TRISB1 = 0;     // output (XC8)

    // RC7 (SDO->SD/DI)
    tris = get_tris_c();
    tris &= ~0x80;
    set_tris_c(tris);
    //TRISC7 = 0;     // output (XC8)

}

/*---------------------*/
/*    Disable SPI      */
/*---------------------*/
static
void power_off (void)	/* Disable MMC/SDC interface */
{
  	// PCH
    setup_spi(SPI_DISABLED);
    //SSPCON1bits.SSPEN = 0;      // MSSP disabled (XC8)
}

/*---------------------*/
/* Send/Receive a byte */
/*---------------------*/
static
FFBYTE xchg_spi (
    FFBYTE dat	/* Data to send */
)
{
  	// PCH
    return spi_read(dat);

    /* XC8
    SSPBUF = dat;
    while (SSPSTATbits.BF != 1);
    return SSPBUF;
    */
}

/*---------------------*/
/* Send multiple bytes */
/*---------------------*/
static
void xmit_spi_multi (
	const FFBYTE *buff,	/* Pointer to the data */
	UINT btx			/* Number of bytes to send (multiple of 4) */
)
{
	do {
		// PCH
		spi_write(*buff++);
	} while (--btx);
	
//	const DWORD *lp = (const DWORD*)buff;
//	RSPI.SPCMD0.BIT.SPB = 3;	/* Set 32-bit mode */
//	do {
//		RSPI.SPDR.LONG = LDDW(*lp++);	/* Send four data bytes */
//		while (!RSPI.SPSR.BIT.SPRF) ;	/* Wait for end of transfer */
//		RSPI.SPDR.LONG;					/* Discard four received bytes */
//	} while (btx -= 4);					/* Repeat until all data sent */
//	RSPI.SPCMD0.BIT.SPB = 7;	/* Set 8-bit mode */
}


/*-----------------------------------------------------------------------*/
/* Deselect card and release SPI                                         */
/*-----------------------------------------------------------------------*/
static
void deselect (void)
{
    CS_HIGH();		/* CS = H */
    xchg_spi(0xFF);	/* Dummy clock (force DO hi-z) */
}

/*------------------------*/
/* Receive multiple bytes */
/*------------------------*/
static
void rcvr_spi_multi (
    FFBYTE *buff,		/* Pointer to data buffer */
    UINT btr		/* Number of bytes to receive (multiple of 4) */
)
{
    do {
    	// PCH
        *buff++ = spi_read(0xFF);

        /* XC8
        SSPBUF = 0xFF;
        while (SSPSTATbits.BF != 1);
        *buff++ = SSPBUF;
        */
    } while (--btr);
    
// Codes below is for RX.
//    DWORD *lp = (DWORD*)buff;
// 	RSPI.SPCMD0.BIT.SPB = 3;	/* Set 32-bit mode */
//    do {
//            RSPI.SPDR.LONG = 0xFFFFFFFF;	/* Send four 0xFFs */
//            while (!RSPI.SPSR.BIT.SPRF) ;	/* Wait for end of transfer */
//            *lp++ = LDDW(RSPI.SPDR.LONG);	/* Store four received bytes */
//    } while (btr -= 4);					/* Repeat until all data received */
//      RSPI.SPCMD0.BIT.SPB = 7;	/* Set 8-bit mode */
}


/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/
static
int wait_ready (    /* 1:Ready, 0:Timeout */
    UINT wt         /* Timeout [ms] */
)
{
    Timer2 = (WORD)wt;

    xchg_spi(0xFF);		/* Read a byte (Force enable DO output) */
    do {
        if (xchg_spi(0xFF) == 0xFF) return 1;	/* Card goes ready */
    } while (Timer2);	/* Wait until card goes ready or timeout */

    return 0;	/* Timeout occured */
}

/*-----------------------------------------------------------------------*/
/* Select card and wait for ready                                        */
/*-----------------------------------------------------------------------*/
static
int select (void)	/* 1:OK, 0:Timeout */
{
    CS_LOW();		/* CS = H */
    if (!wait_ready(500)) {
            deselect();
            return 0;	/* Failed to select the card due to timeout */
    }
    return 1;	/* OK */
}

/*-----------------------------------------------------------------------*/
/* Receive a data packet from the MMC                                    */
/*-----------------------------------------------------------------------*/
static
int rcvr_datablock (        /* 1:OK, 0:Error */
    FFBYTE *buff,			/* Data buffer */
    UINT btr			/* Data block length (byte) */
)
{
    FFBYTE token;

    Timer1 = 200;
    do {				/* Wait for DataStart token in timeout of 200ms */
        token = xchg_spi(0xFF);
        /* This loop will take a time. Insert rot_rdq() here for multitask envilonment. */
    } while ((token == 0xFF) && Timer1);
    if (token != 0xFE) return 0;	/* Function fails if invalid DataStart token or timeout */

    rcvr_spi_multi(buff, btr);		/* Store trailing data to the buffer */
    xchg_spi(0xFF); xchg_spi(0xFF);	/* Discard CRC */

    return 1;						/* Function succeeded */
}


// XC8 does not allow recursive call.
static
FFBYTE send_cmd_CMD55 (		/* Return value: R1 resp (bit7==1:Failed to send) */
    FFBYTE cmd,		/* Command index */
    DWORD arg		/* Argument */
)
{
    FFBYTE n, res;

//    if (cmd & 0x80) {	/* Send a CMD55 prior to ACMD<n> */
//        cmd &= 0x7F;
//        res = send_cmd(CMD55, 0);
//        if (res > 1) return res;
//    }

    /* Select the card and wait for ready except to stop multiple block read */
    if (cmd != CMD12) {
        deselect();
        if (!select()) return 0xFF;
    }

    /* Send command packet */
    xchg_spi(0x40 | cmd);		/* Start + command index */
    xchg_spi((FFBYTE)(arg >> 24));	/* Argument[31..24] */
    xchg_spi((FFBYTE)(arg >> 16));	/* Argument[23..16] */
    xchg_spi((FFBYTE)(arg >> 8));		/* Argument[15..8] */
    xchg_spi((FFBYTE)arg);		/* Argument[7..0] */
    n = 0x01;				/* Dummy CRC + Stop */
    if (cmd == CMD0) n = 0x95;		/* Valid CRC for CMD0(0) */
    if (cmd == CMD8) n = 0x87;		/* Valid CRC for CMD8(0x1AA) */
    xchg_spi(n);

    /* Receive command resp */
    if (cmd == CMD12) xchg_spi(0xFF);	/* Diacard stuff byte on CMD12 */
    n = 10;				/* Wait for response (10 bytes max) */
    do
        res = xchg_spi(0xFF);
    while ((res & 0x80) && --n);

    return res;				/* Return received response */
}


/*-----------------------------------------------------------------------*/
/* Send a command packet to the MMC                                      */
/*-----------------------------------------------------------------------*/
static
FFBYTE send_cmd (		/* Return value: R1 resp (bit7==1:Failed to send) */
    FFBYTE cmd,		/* Command index */
    DWORD arg		/* Argument */
)
{
    FFBYTE n, res;

    if (cmd & 0x80) {	/* Send a CMD55 prior to ACMD<n> */
        cmd &= 0x7F;
        res = send_cmd_CMD55(CMD55, 0);
        if (res > 1) return res;
    }

    /* Select the card and wait for ready except to stop multiple block read */
    if (cmd != CMD12) {
        deselect();
        if (!select()) return 0xFF;
    }

    /* Send command packet */
    xchg_spi(0x40 | cmd);		/* Start + command index */
    xchg_spi((FFBYTE)(arg >> 24));	/* Argument[31..24] */
    xchg_spi((FFBYTE)(arg >> 16));	/* Argument[23..16] */
    xchg_spi((FFBYTE)(arg >> 8));		/* Argument[15..8] */
    xchg_spi((FFBYTE)arg);		/* Argument[7..0] */
    n = 0x01;				/* Dummy CRC + Stop */
    if (cmd == CMD0) n = 0x95;		/* Valid CRC for CMD0(0) */
    if (cmd == CMD8) n = 0x87;		/* Valid CRC for CMD8(0x1AA) */
    xchg_spi(n);

    /* Receive command resp */
    if (cmd == CMD12) xchg_spi(0xFF);	/* Diacard stuff byte on CMD12 */
    n = 10;				/* Wait for response (10 bytes max) */
    do
        res = xchg_spi(0xFF);
    while ((res & 0x80) && --n);

    return res;				/* Return received response */
}

/*-----------------------------------------------------------------------*/
/* Inidialize a Drive                                                    */
/*-----------------------------------------------------------------------*/
DSTATUS disk_initialize (
    FFBYTE drv				/* Physical drive nmuber (0..) */
)
{
    FFBYTE n, cmd, ty, ocr[4];

    if (drv) return STA_NOINIT;			/* Supports only drive 0 */

    power_on();							/* Initialize RSPI */
    for (n = 10; n; n--) xchg_spi(0xFF);/* Send 80 dummy clocks */

    ty = 0;
    if (send_cmd(CMD0, 0) == 1) {                               /* Put the card SPI/Idle state */
        Timer1 = 1000;						/* Initialization timeout = 1 sec */
        if (send_cmd(CMD8, 0x1AA) == 1) {                       /* SDv2? */
            for (n = 0; n < 4; n++) ocr[n] = xchg_spi(0xFF);	/* Get 32 bit return value of R7 resp */
            if (ocr[2] == 0x01 && ocr[3] == 0xAA) {		/* Is the card supports vcc of 2.7-3.6V? */
                while (Timer1 && send_cmd(ACMD41, 1UL << 30)) ;	/* Wait for end of initialization with ACMD41(HCS) */
                if (Timer1 && send_cmd(CMD58, 0) == 0) {	/* Check CCS bit in the OCR */
                    for (n = 0; n < 4; n++) ocr[n] = xchg_spi(0xFF);
                    ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;	/* Card id SDv2 */
                }
            }
        } else {	/* Not SDv2 card */
            if (send_cmd(ACMD41, 0) <= 1) 	{	/* SDv1 or MMC? */
                ty = CT_SD1; cmd = ACMD41;              /* SDv1 (ACMD41(0)) */
            } else {
                ty = CT_MMC; cmd = CMD1;                /* MMCv3 (CMD1(0)) */
            }
            while (Timer1 && send_cmd(cmd, 0)) ;	/* Wait for end of initialization */
            if (!Timer1 || send_cmd(CMD16, 512) != 0)	/* Set block length: 512 */
                ty = 0;
        }
    }
    CardType = ty;	/* Card type */
    deselect();

    if (ty) {			/* OK */
//TODO:for PIC?            FCLK_FAST();			/* Set fast clock */
        Stat &= ~STA_NOINIT;	/* Clear STA_NOINIT flag */
    } else {			/* Failed */
        power_off();
        Stat = STA_NOINIT;
    }

    return Stat;
}


/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/
DSTATUS disk_status (
    FFBYTE drv		/* Physical drive nmuber (0..) */
)
{
    if (drv) return STA_NOINIT;		/* Supports only drive 0 */

    return Stat;	/* Return disk status */
}


/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/
DRESULT disk_read (
    FFBYTE drv,		/* Physical drive nmuber (0..) */
    FFBYTE *buff,		/* Data buffer to store read data */
    DWORD sector,           /* Sector address (LBA) */
    UINT count		/* Number of sectors to read (1..128) */
)
{
    FFBYTE cmd;

    if (drv || !count) return RES_PARERR;		/* Check parameter */
    if (Stat & STA_NOINIT) return RES_NOTRDY;           /* Check if drive is ready */
    if (!(CardType & CT_BLOCK)) sector *= 512;          /* LBA ot BA conversion (byte addressing cards) */

    cmd = count > 1 ? CMD18 : CMD17;			/*  READ_MULTIPLE_BLOCK : READ_SINGLE_BLOCK */
    if (send_cmd(cmd, sector) == 0) {
        do {
            if (!rcvr_datablock(buff, 512)) break;
            buff += 512;
        } while (--count);
        if (cmd == CMD18) send_cmd(CMD12, 0);           /* STOP_TRANSMISSION */
    }
    deselect();

    return count ? RES_ERROR : RES_OK;                  /* Return result */
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/
#if _USE_WRITE
static
int xmit_datablock (	/* 1:OK, 0:Failed */
	const FFBYTE *buff,	/* Ponter to 512 byte data to be sent */
	FFBYTE token			/* Token */
)
{
	FFBYTE resp;


	if (!wait_ready(500)) return 0;		/* Wait for card ready */

	xchg_spi(token);					/* Send token */
	if (token != 0xFD) {				/* Send data if token is other than StopTran */
		xmit_spi_multi(buff, 512);		/* Data */
		xchg_spi(0xFF); xchg_spi(0xFF);	/* Dummy CRC */

		resp = xchg_spi(0xFF);			/* Receive data resp */
		if ((resp & 0x1F) != 0x05)		/* Function fails if the data packet was not accepted */
			return 0;
	}
	return 1;
}

DRESULT disk_write (
    FFBYTE drv,			/* Physical drive nmuber (0..) */
    const FFBYTE *buff,	/* Data to be written */
    DWORD sector,		/* Sector address (LBA) */
    UINT count			/* Number of sectors to write (1..128) */
)
{
	if (drv || !count) return RES_PARERR;		/* Check parameter */
	if (Stat & STA_NOINIT) return RES_NOTRDY;	/* Check drive status */
	if (Stat & STA_PROTECT) return RES_WRPRT;	/* Check write protect */

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* LBA ==> BA conversion (byte addressing cards) */

	if (count == 1) {	/* Single sector write */
		if ((send_cmd(CMD24, sector) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE))
			count = 0;
	}
	else {				/* Multiple sector write */
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sector) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD))	/* STOP_TRAN token */
				count = 1;
		}
	}
	deselect();

	return count ? RES_ERROR : RES_OK;	/* Return result */
}
#endif


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL
DRESULT disk_ioctl (
    FFBYTE pdrv,		/* Physical drive nmuber (0..) */
    FFBYTE cmd,		/* Control code */
    void *buff		/* Buffer to send/receive control data */
)
{
    /* TODO
    DRESULT res;
    int result;
    */
    
    return RES_PARERR;
}
#endif

DWORD get_fattime (void)
{
	/* No RTC feature provided. Return a fixed value 2015/5/3 0:00:00 */
	return	  ((DWORD)(2015 - 1980) << 25)	/* Y */
			| ((DWORD)5  << 21)				/* M */
			| ((DWORD)3 << 16)				/* D */
			| ((DWORD)0  << 11)				/* H */
			| ((DWORD)0  << 5)				/* M */
			| ((DWORD)0  >> 1);				/* S */
}

